@@ -11,7 +11,7 @@ from struct import error as StructError
 import bz2
 import zlib
 import datetime
-from decimal import localcontext, Decimal
+from decimal import Context
 from uuid import UUID
 
 import json
@@ -26,7 +26,7 @@ from ._schema_common import SCHEMA_DEFS
 from ._read_common import (
     SchemaResolutionError, MAGIC, SYNC_SIZE, HEADER_SCHEMA, missing_codec_lib
 )
-from ._timezone import utc
+from ._timezone import epoch
 from .const import (
     MCS_PER_HOUR, MCS_PER_MINUTE, MCS_PER_SECOND, MLS_PER_HOUR, MLS_PER_MINUTE,
     MLS_PER_SECOND, DAYS_SHIFT
@@ -53,10 +53,14 @@ AVRO_TYPES = {
     'error_union'
 }
 
+decimal_context = Context()
+
 
 def match_types(writer_type, reader_type):
     if isinstance(writer_type, list) or isinstance(reader_type, list):
         return True
+    if isinstance(writer_type, dict) or isinstance(reader_type, dict):
+        return match_schemas(writer_type, reader_type)
     if writer_type == reader_type:
         return True
     # promotion cases
@@ -118,7 +122,7 @@ def read_boolean(decoder, writer_schema=None, reader_schema=None):
 
 
 def parse_timestamp(data, resolution):
-    return datetime.datetime.fromtimestamp(data / resolution, tz=utc)
+    return epoch + datetime.timedelta(seconds=data / resolution)
 
 
 def read_timestamp_millis(data, writer_schema=None, reader_schema=None):
@@ -159,10 +163,9 @@ def read_decimal(data, writer_schema=None, reader_schema=None):
 
     unscaled_datum = be_signed_bytes_to_int(data)
 
-    with localcontext() as ctx:
-        ctx.prec = precision
-        scaled_datum = Decimal(unscaled_datum).scaleb(-scale)
-    return scaled_datum
+    decimal_context.prec = precision
+    return decimal_context.create_decimal(unscaled_datum). \
+        scaleb(-scale, decimal_context)
 
 
 def read_int(decoder, writer_schema=None, reader_schema=None):
@@ -207,33 +210,40 @@ def read_enum(decoder, writer_schema, reader_schema=None):
     return symbol
 
 
-def read_array(decoder, writer_schema, reader_schema=None):
+def read_array(decoder, writer_schema, reader_schema=None,
+               return_record_name=False):
     if reader_schema:
-        def item_reader(decoder, w_schema, r_schema):
-            return read_data(decoder, w_schema['items'], r_schema['items'])
+        def item_reader(decoder, w_schema, r_schema, return_record_name):
+            return read_data(decoder, w_schema['items'], r_schema['items'],
+                             return_record_name)
     else:
-        def item_reader(decoder, w_schema, _):
-            return read_data(decoder, w_schema['items'])
+        def item_reader(decoder, w_schema, _, return_record_name):
+            return read_data(decoder, w_schema['items'], None,
+                             return_record_name)
 
     read_items = []
 
     decoder.read_array_start()
 
     for item in decoder.iter_array():
-        read_items.append(item_reader(decoder, writer_schema, reader_schema))
+        read_items.append(item_reader(decoder, writer_schema, reader_schema,
+                                      return_record_name))
 
     decoder.read_array_end()
 
     return read_items
 
 
-def read_map(decoder, writer_schema, reader_schema=None):
+def read_map(decoder, writer_schema, reader_schema=None,
+             return_record_name=False):
     if reader_schema:
         def item_reader(decoder, w_schema, r_schema):
-            return read_data(decoder, w_schema['values'], r_schema['values'])
+            return read_data(decoder, w_schema['values'], r_schema['values'],
+                             return_record_name)
     else:
         def item_reader(decoder, w_schema, _):
-            return read_data(decoder, w_schema['values'])
+            return read_data(decoder, w_schema['values'], None,
+                             return_record_name)
 
     read_items = {}
 
@@ -248,26 +258,50 @@ def read_map(decoder, writer_schema, reader_schema=None):
     return read_items
 
 
-def read_union(decoder, writer_schema, reader_schema=None):
+def read_union(decoder, writer_schema, reader_schema=None,
+               return_record_name=False):
     # schema resolution
     index = decoder.read_index()
+    idx_schema = writer_schema[index]
     if reader_schema:
         # Handle case where the reader schema is just a single type (not union)
         if not isinstance(reader_schema, list):
-            if match_types(writer_schema[index], reader_schema):
-                return read_data(decoder, writer_schema[index], reader_schema)
+            if match_types(idx_schema, reader_schema):
+                return read_data(
+                    decoder, idx_schema, reader_schema, return_record_name
+                )
         else:
             for schema in reader_schema:
-                if match_types(writer_schema[index], schema):
-                    return read_data(decoder, writer_schema[index], schema)
+                if match_types(idx_schema, schema):
+                    return read_data(
+                        decoder, idx_schema, schema, return_record_name
+                    )
         msg = 'schema mismatch: %s not found in %s' % \
             (writer_schema, reader_schema)
         raise SchemaResolutionError(msg)
     else:
-        return read_data(decoder, writer_schema[index])
+        if (
+            return_record_name and extract_record_type(idx_schema) == 'record'
+        ):
+            return (
+                idx_schema['name'],
+                read_data(decoder, idx_schema, None, return_record_name)
+            )
+        elif (
+            return_record_name
+            and extract_record_type(idx_schema) not in AVRO_TYPES
+        ):
+            # idx_schema is a named type
+            return (
+                SCHEMA_DEFS[idx_schema]['name'],
+                read_data(decoder, idx_schema, None, return_record_name)
+            )
+        else:
+            return read_data(decoder, idx_schema)
 
 
-def read_record(decoder, writer_schema, reader_schema=None):
+def read_record(decoder, writer_schema, reader_schema=None,
+                return_record_name=False):
     """A record is encoded by encoding the values of its fields in the order
     that they are declared. In other words, a record is encoded as just the
     concatenation of the encodings of its fields.  Field values are encoded per
@@ -289,7 +323,8 @@ def read_record(decoder, writer_schema, reader_schema=None):
     record = {}
     if reader_schema is None:
         for field in writer_schema['fields']:
-            record[field['name']] = read_data(decoder, field['type'])
+            record[field['name']] = read_data(decoder, field['type'], None,
+                                              return_record_name)
     else:
         readers_field_dict = {}
         aliases_field_dict = {}
@@ -308,10 +343,11 @@ def read_record(decoder, writer_schema, reader_schema=None):
                     decoder,
                     field['type'],
                     readers_field['type'],
+                    return_record_name,
                 )
             else:
                 # should implement skip
-                read_data(decoder, field['type'], field['type'])
+                read_data(decoder, field['type'], None, return_record_name)
 
         # fill in default values
         if len(readers_field_dict) > len(record):
@@ -375,11 +411,11 @@ def maybe_promote(data, writer_type, reader_type):
     return data
 
 
-def read_data(decoder, writer_schema, reader_schema=None):
+def read_data(decoder, writer_schema, reader_schema=None,
+              return_record_name=False):
     """Read data from file object according to schema."""
 
     record_type = extract_record_type(writer_schema)
-    logical_type = extract_logical_type(writer_schema)
 
     if reader_schema and record_type in AVRO_TYPES:
         # If the schemas are the same, set the reader schema to None so that no
@@ -392,13 +428,18 @@ def read_data(decoder, writer_schema, reader_schema=None):
     reader_fn = READERS.get(record_type)
     if reader_fn:
         try:
-            data = reader_fn(decoder, writer_schema, reader_schema)
+            if record_type in ['array', 'map', 'record', 'union']:
+                data = reader_fn(decoder, writer_schema, reader_schema,
+                                 return_record_name)
+            else:
+                data = reader_fn(decoder, writer_schema, reader_schema)
         except StructError:
             raise EOFError(
                 'cannot read %s from %s' % (record_type, decoder.fo)
             )
 
         if 'logicalType' in writer_schema:
+            logical_type = extract_logical_type(writer_schema)
             fn = LOGICAL_READERS.get(logical_type)
             if fn:
                 return fn(data, writer_schema, reader_schema)
@@ -415,7 +456,8 @@ def read_data(decoder, writer_schema, reader_schema=None):
         return read_data(
             decoder,
             SCHEMA_DEFS[record_type],
-            SCHEMA_DEFS.get(reader_schema)
+            SCHEMA_DEFS.get(reader_schema),
+            return_record_name
         )
 
 
@@ -480,7 +522,22 @@ else:
     BLOCK_READERS["zstandard"] = zstandard_read_block
 
 
-def _iter_avro_records(decoder, header, codec, writer_schema, reader_schema):
+def lz4_read_block(decoder):
+    length = read_long(decoder)
+    data = decoder.read_fixed(length)
+    return MemoryIO(lz4.block.decompress(data))
+
+
+try:
+    import lz4.block
+except ImportError:
+    BLOCK_READERS["lz4"] = missing_codec_lib("lz4", "lz4")
+else:
+    BLOCK_READERS["lz4"] = lz4_read_block
+
+
+def _iter_avro_records(decoder, header, codec, writer_schema, reader_schema,
+                       return_record_name=False):
     """Return iterator over avro records."""
     sync_marker = header['sync']
 
@@ -499,13 +556,15 @@ def _iter_avro_records(decoder, header, codec, writer_schema, reader_schema):
 
         for i in xrange(block_count):
             yield read_data(
-                BinaryDecoder(block_fo), writer_schema, reader_schema
+                BinaryDecoder(block_fo), writer_schema, reader_schema,
+                return_record_name
             )
 
         skip_sync(decoder.fo, sync_marker)
 
 
-def _iter_avro_blocks(decoder, header, codec, writer_schema, reader_schema):
+def _iter_avro_blocks(decoder, header, codec, writer_schema, reader_schema,
+                      return_record_name=False):
     """Return iterator over avro blocks."""
     sync_marker = header['sync']
 
@@ -528,7 +587,7 @@ def _iter_avro_blocks(decoder, header, codec, writer_schema, reader_schema):
 
         yield Block(
             block_bytes, num_block_records, codec, reader_schema,
-            writer_schema, offset, size
+            writer_schema, offset, size, return_record_name
         )
 
 
@@ -555,8 +614,9 @@ class Block:
 
         Size of the block in bytes
     """
+
     def __init__(self, bytes_, num_records, codec, reader_schema,
-                 writer_schema, offset, size):
+                 writer_schema, offset, size, return_record_name=False):
         self.bytes_ = bytes_
         self.num_records = num_records
         self.codec = codec
@@ -564,6 +624,7 @@ class Block:
         self.writer_schema = writer_schema
         self.offset = offset
         self.size = size
+        self.return_record_name = return_record_name
 
     def __iter__(self):
         for i in xrange(self.num_records):
@@ -571,6 +632,7 @@ class Block:
                 BinaryDecoder(self.bytes_),
                 self.writer_schema,
                 self.reader_schema,
+                self.return_record_name
             )
 
     def __str__(self):
@@ -580,7 +642,8 @@ class Block:
 
 
 class file_reader(object):
-    def __init__(self, fo_or_decoder, reader_schema=None):
+    def __init__(self, fo_or_decoder, reader_schema=None,
+                 return_record_name=False):
         if isinstance(fo_or_decoder, AvroJSONDecoder):
             self.decoder = fo_or_decoder
         else:
@@ -593,11 +656,13 @@ class file_reader(object):
             )
         else:
             self.reader_schema = None
+        self.return_record_name = return_record_name
         self._elems = None
 
     def _read_header(self):
         try:
-            self._header = read_data(self.decoder, HEADER_SCHEMA)
+            self._header = read_data(self.decoder, HEADER_SCHEMA, None,
+                                     self.return_record_name)
         except StopIteration:
             raise ValueError('cannot read header - is it an avro file?')
 
@@ -644,6 +709,10 @@ class reader(file_reader):
         Input stream
     reader_schema: dict, optional
         Reader schema
+    return_record_name: bool, optional
+        If true, when reading a union of records, the result will be a tuple
+        where the first value is the name of the record and the second value is
+        the record itself
 
 
     Example::
@@ -654,6 +723,18 @@ class reader(file_reader):
             for record in avro_reader:
                 process_record(record)
 
+    The `fo` argument is a file-like object so another common example usage
+    would use an `io.BytesIO` object like so::
+
+        from io import BytesIO
+        from fastavro import writer, reader
+
+        fo = BytesIO()
+        writer(fo, schema, records)
+        fo.seek(0)
+        for record in reader(fo):
+            process_record(record)
+
     .. attribute:: metadata
 
         Key-value pairs in the header metadata
@@ -671,8 +752,8 @@ class reader(file_reader):
         The schema used when reading (if provided)
     """
 
-    def __init__(self, fo, reader_schema=None):
-        file_reader.__init__(self, fo, reader_schema)
+    def __init__(self, fo, reader_schema=None, return_record_name=False):
+        file_reader.__init__(self, fo, reader_schema, return_record_name)
 
         if isinstance(self.decoder, AvroJSONDecoder):
             self.decoder.configure(self.reader_schema)
@@ -686,8 +767,10 @@ class reader(file_reader):
                         self.decoder,
                         self.writer_schema,
                         self.reader_schema,
+                        self.return_record_name
                     )
                     self.decoder.drain()
+
             self._elems = _elems()
 
         else:
@@ -697,7 +780,8 @@ class reader(file_reader):
                                              self._header,
                                              self.codec,
                                              self.writer_schema,
-                                             self.reader_schema)
+                                             self.reader_schema,
+                                             self.return_record_name)
 
 
 class block_reader(file_reader):
@@ -709,6 +793,10 @@ class block_reader(file_reader):
         Input stream
     reader_schema: dict, optional
         Reader schema
+    return_record_name: bool, optional
+        If true, when reading a union of records, the result will be a tuple
+        where the first value is the name of the record and the second value is
+        the record itself
 
 
     Example::
@@ -736,8 +824,8 @@ class block_reader(file_reader):
         The schema used when reading (if provided)
     """
 
-    def __init__(self, fo, reader_schema=None):
-        file_reader.__init__(self, fo, reader_schema)
+    def __init__(self, fo, reader_schema=None, return_record_name=False):
+        file_reader.__init__(self, fo, reader_schema, return_record_name)
 
         self._read_header()
 
@@ -745,10 +833,12 @@ class block_reader(file_reader):
                                         self._header,
                                         self.codec,
                                         self.writer_schema,
-                                        self.reader_schema)
+                                        self.reader_schema,
+                                        self.return_record_name)
 
 
-def schemaless_reader(fo, writer_schema, reader_schema=None):
+def schemaless_reader(fo, writer_schema, reader_schema=None,
+                      return_record_name=False):
     """Reads a single record writen using the
     :meth:`~fastavro._write_py.schemaless_writer`
 
@@ -761,6 +851,10 @@ def schemaless_reader(fo, writer_schema, reader_schema=None):
     reader_schema: dict, optional
         If the schema has changed since being written then the new schema can
         be given to allow for schema migration
+    return_record_name: bool, optional
+        If true, when reading a union of records, the result will be a tuple
+        where the first value is the name of the record and the second value is
+        the record itself
 
 
     Example::
@@ -785,7 +879,8 @@ def schemaless_reader(fo, writer_schema, reader_schema=None):
     return read_data(
         decoder,
         writer_schema,
-        reader_schema
+        reader_schema,
+        return_record_name,
     )
 
 
